# graphene imports
import graphene
# error management
from django_graphbox.exceptions import ErrorMsgType

# login mutate function builder

def build_mutate_for_login(self):
    def login_mutate_function(parent, info, **kwargs):
        config=self._models_by_op_name['login']
        login_id_value=kwargs.get(self._session_manager.login_id_field_name)
        password_value=kwargs.get(self._session_manager.password_field_name)
        permanent=kwargs.get('permanent')
        valid, user_instance, token, error=self._session_manager.start_session(login_id_value, password_value, permanent)
        return_object=type(info.return_type.name, (graphene.ObjectType,), {'estado':graphene.Boolean(),'token':graphene.String(), user_instance.__class__.__name__:graphene.Field(config['type']),'error':graphene.Field(ErrorMsgType)})
        return return_object(**{'estado':valid,'token':token, user_instance.__class__.__name__:user_instance, 'error':error})
    return login_mutate_function

# social login mutate function builder
def build_mutate_for_social_login(self):
    def social_login_mutate_function(parent, info, **kwargs):
        config=self._models_by_op_name['login']
        token=kwargs.get('token')
        origin=kwargs.get('origin')
        valid, user_instance, token, error=self._session_manager.start_social_session(token, origin)
        return_object=type(info.return_type.name, (graphene.ObjectType,), {'estado':graphene.Boolean(),'token':graphene.String(), user_instance.__class__.__name__:graphene.Field(config['type']),'error':graphene.Field(ErrorMsgType)})
        return return_object(**{'estado':valid,'token':token, user_instance.__class__.__name__:user_instance, 'error':error})
    return social_login_mutate_function

# actual user query resolver builder

def build_actual_user_resolver(self):
    def actual_user_function(parent, info, **kwargs):
        valid, user_instance, error=self._session_manager.validate_access(info.context, 'all')
        return user_instance
    return actual_user_function


""" Error responses for GraphBox API """
from graphene_django.types import ObjectType
import graphene
from .constants import *

class ErrorMsgType(ObjectType):
    """Custom ObjectType to manage the exceptions of the django_graphbox schema.
    """
    codigo=graphene.Int()
    message=graphene.String()
    description=graphene.String()

class ErrorManager:
    """Class to manage the exceptions of the django_graphbox schema.
    """
    _error_list={
        NO_ERROR: {
            "message": "Operación ejeccutada correctamente",
            "description": "No ha ocurrido ningún error"
        },
        UNKNOWN_ERROR: {
            "message": "Error desconocido",
            "description": "Ha ocurrido un error desconocido"
        },
        INVALID_CREDENTIALS: {
            "message": "Credenciales inválidas",
            "description": "Las credenciales proporcionadas no son válidas"
        },
        INVALID_TOKEN: {
            "message": "Token inválido",
            "description": "El token proporcionado no es válido"
        },
        EXPIRED_TOKEN: {
            "message": "Token expirado",
            "description": "El token proporcionado ha expirado"
        },
        INTERNAL_ERROR: {
            "message": "Error interno",
            "description": "Ha ocurrido un error interno"
        },
        ACCESS_DENIED: {
            "message": "Acceso denegado",
            "description": "No tiene permisos para realizar esta acción"
        },
        INSTANCE_NOT_FOUND: {
            "message": "Instancia no encontrada",
            "description": "No se ha encontrado la instancia solicitada"
        },
        INSUFFICIENT_PERMISSIONS: {
            "message": "Permisos insuficientes",
            "description": "No tiene permisos suficientes para realizar esta acción"
        },
        USER_ALREADY_EXISTS: {
            "message": "El usuario ya esta en uso.",
            "description": "Esta cuenta de usuario ya fue registrada."
        }
    }

    @classmethod
    def get_error_by_code(cls, error_code=None, custom_message=None, custom_description=None):
        """Method to get the error message by code.

        Args:
            error_code (int): Code of the error.
            custom_message (str): Custom message to replace the default message.
            custom_description (str): Custom description to replace the default description.
            
        Returns:
            ErrorMsgType: Error message.
        """
        if error_code is None:
            error_code=UNKNOWN_ERROR
        if custom_message != None and custom_description != None:
            return ErrorMsgType(codigo=error_code, message=custom_message, description=custom_description)
        else:
            if error_code not in cls._error_list.keys():
                error_code=UNKNOWN_ERROR
            return ErrorMsgType(codigo=error_code, message=cls._error_list[error_code]['message'], description=cls._error_list[error_code]['description'])


""" This module implements a Session Manager JWT based, that can be used on SchemaBuilder for authentication.
"""
# jwt
import jwt
# settings import
from django.conf import settings
# randomic session id generator
from uuid import uuid4
# time management import
from django.utils import timezone as tz
import datetime
# django password validator
from django.contrib.auth.hashers import check_password
# error manager
from .exceptions import ErrorManager
# global constants
from .constants import *
# google imports
from google.oauth2 import id_token
from google.auth.transport import requests as grq
# facebook imports
import facebook
# files manage
from django.core.files.images import ImageFile
from django.core.files.temp import tempfile
import requests
from .hasher import HashManager

class GroupManager:
    """ Manager for allow access to users by groups. """
    _groups={}
    _modify_permissions={}
    _user_model=None
    _rol_field_name='rol'

    def __init__(self, user_model, rol_field_name, groups={}, modify_permissions={}):
        """ Initialize the GroupManager.
        
        Args:
            user_model (class): The user model class.
            rol_field_name (str): The name of the field that contains the user rol.
            groups (dict): The groups and their user rol list. {'group_name': ['rol_1', 'rol_2', ...], ...}
            modify_permissions (dict): The user rol and their modify permissions for change data of other users. {'rol_1': ['rol_1', 'rol_2', ...], ...}
        """
        self._user_model=user_model
        self._rol_field_name=rol_field_name
        self._groups=groups
        self._modify_permissions=modify_permissions

    def add_group(self, group_name, access_list):
        """ Add a new group to the manager. """
        self._groups[group_name]=access_list

    def validar_acesso(self, user_instance, group_name):
        """ Validate if a user_instance is on a group_name. """
        if self._user_model!=None and isinstance(user_instance, self._user_model):
            if hasattr(user_instance, self._rol_field_name):
                if group_name=='all' or getattr(user_instance, self._rol_field_name) in self._groups[group_name]:
                    return True
                return False
    
    def get_modify_permissions(self, user_instance):
        """ Get the modify permissions for a user_instance. """
        if user_instance.rol in self._modify_permissions:
            return self._modify_permissions[user_instance.rol]
        return []

class GoogleSession:
    client_id=getattr(settings, 'GOOGLE_CLIENT_ID', None)
    @classmethod
    def validate (cls, token):
        try:
            info=id_token.verify_oauth2_token(token, grq.Request(), cls.client_id)
            return True, info
        except:
            data=[]
            return False,data

class FacebookSession:
    app_id=getattr(settings, 'FACEBOOK_APP_ID', None)
    @classmethod
    def validate (cls, token):
        try:
            graph=facebook.GraphAPI(access_token=token, version='3.1')
            info=graph.get_object('me', fields='id,email,name,picture')
            return True, info
        except:
            data=[]
            return False,data

class Manager:
    _security_key=settings.SECRET_KEY
    social_origin_facebook='FACEBOOK'
    social_origin_google='GOOGLE'

    def __init__(self, user_model, use_social_session=False, rol_field_name='rol', login_id_field_name='uname', password_field_name='password', name_field_name="name", photo_field_name="photo", social_id_field=None, social_origin_field=None, active_field_name=None, session_expiration_time=12, groups={}, modify_permissions={}, security_key=None):
        """Session Manager for a user_model

        Args:
            user_model (django.models.Model): User model
            use_social_session (bool, optional): Use social session. Defaults to False.
            rol_field_name (str, optional): Field name of user model for access level control. Defaults to 'rol'.
            login_id_field_name (str, optional): Field name of unique field used for authentication. Defaults to 'uname'.
            password_field_name (str, optional): Field name of password field used for authentication. Defaults to 'password'.
            name_field_name (str, optional): Field name of name field used for authentication. Defaults to 'name'. (Only for social login)              
            photo_field_name (str, optional): Field name of photo field used for authentication. Defaults to 'photo'. (Only for social login)
            social_id_field (str, optional): Field name of social id field used for authentication. Defaults to None. (Only for social login)
            social_origin_field (str, optional): Field name of social origin field used for authentication. Defaults to None. (Only for social login)
            active_field_name (str, optional): Field name for boolean field name for validation active user. Defaults to None for not validation.
            session_expiration_time (int, optional): Time in hours to set expiration time on jwt. Defaults to 12.
            groups (dict, optional): Predefined access groups {"group_name":["rol_1", "rol_2", "rol_3"]}. Defaults to {}.
            modify_permissions (dict, optional): Predefined modify permissions {"rol_1":["rol_1", "rol_2", "rol_3"]}. Defaults to {}.
            security_key (str, optional): Secret used as secret for sign on jwt. Defaults to None for use django.conf.settings.SECRET_KEY.
        """
        id_attribute=getattr(user_model, login_id_field_name)
        if id_attribute.field.unique==False:
            raise Exception('Field '+login_id_field_name+' must be unique')
        self.user_model=user_model
        self.use_social_session=use_social_session
        self.rol_field_name=rol_field_name
        self.login_id_field_name=login_id_field_name
        self.password_field_name=password_field_name
        self.name_field_name=name_field_name
        self.photo_field_name=photo_field_name
        self.social_id_field=social_id_field
        self.social_origin_field=social_origin_field
        self.active_field_name=active_field_name
        self.session_expiration_time=session_expiration_time
        self.group_manager=GroupManager(user_model, rol_field_name, groups, modify_permissions)
        if security_key!=None:
            self._security_key=security_key
    
    def _generate_token(self, user_instance, expiration_time=0):
        """Generate jwt for user_instance

        Args:
            user_instance (user_model.__class__): user instance to generate token
            expiration_time (int, optional): expiration time in hours. Defaults to 0 for no expiration.
        Returns:
            str: generated token
        """
        if self.user_model!=None and isinstance(user_instance, self.user_model):
            if hasattr(user_instance, self.rol_field_name):
                session_id=str(uuid4())
                payload={
                    'session_id':session_id,
                    'u_id': user_instance.id,
                }
                if expiration_time>0:
                    payload['exp']=tz.localtime()+datetime.timedelta(hours=expiration_time)
                return jwt.encode(payload, self._security_key, algorithm='HS256')
        return None
    
    def validate_access(self, request, group_name):
        """Validate access

        Args:
            request (django.http.request.HttpRequest): request to validate Authorization header as Bearer token
            group_name (str): group name to validate
        Returns:
            tuple:(status (bool), user_instance (UserObject))
        """
        if group_name=='open' or group_name==None:
            return True, None, ErrorManager.get_error_by_code(NO_ERROR)
        if 'Authorization' in request.headers:
            token=request.headers['Authorization']
            token=token[7:len(token)]
            try:
                payload=jwt.decode(token, self._security_key, algorithms=['HS256'])
                if self.user_model.objects.filter(id=payload['u_id']).exists():
                    user_instance=self.user_model.objects.get(id=payload['u_id'])
                    if self.group_manager.validar_acesso(user_instance, group_name):
                        return True, user_instance, ErrorManager.get_error_by_code(NO_ERROR)
                    else:
                        return False, None, ErrorManager.get_error_by_code(ACCESS_DENIED)
                else:
                    return False, None, ErrorManager.get_error_by_code(INVALID_CREDENTIALS)
            except:
                return False, None, ErrorManager.get_error_by_code(INVALID_TOKEN)
        else:
            return False, None, ErrorManager.get_error_by_code(INVALID_TOKEN)

    def start_session(self, login_id_value, password, permanent=False):
        """ Start session

        Args:
            login_id_value (str): login id value
            password (str): password
            permanent (bool, optional): if True, session will be permanent. Defaults to False.
        Returns:
            tuple:(status (bool), user_insrance (UserObject), token (str), error_message (ErrorMsgType))
        """
        if self.user_model!=None:
            try:
                if self.user_model.objects.filter(**{self.login_id_field_name:login_id_value}).exists():
                    user_instance=self.user_model.objects.get(**{self.login_id_field_name:login_id_value})
                    if self.active_field_name==None or getattr(user_instance, self.active_field_name):
                        if check_password(password, getattr(user_instance, self.password_field_name)):
                            token=self._generate_token(user_instance, 0 if permanent else self.session_expiration_time)
                            if token!=None:
                                return True, user_instance, token, ErrorManager.get_error_by_code(NO_ERROR)
                            else:
                                return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
                        else:
                            return False, None, '', ErrorManager.get_error_by_code(INVALID_CREDENTIALS)
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(INVALID_CREDENTIALS)
                else:
                    return False, None, '', ErrorManager.get_error_by_code(INVALID_CREDENTIALS)
            except:
                return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
        else:
            return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)

    def download_photo(self, url):
        img_temp = tempfile.TemporaryFile(suffix='.jpg')
        r=requests.get(url)
        img_temp.write(r.content)
        img_temp.flush()
        photo=ImageFile(img_temp)
        return photo

    def start_social_session(self, token, origin):
        """ Start session from social network

        Args:
            token (str): token
            origin (str): origin
        Returns:
            tuple:(status (bool), user_insrance (UserObject), token (str), error_message (ErrorMsgType))
        """
        if self.user_model!=None:
            try:
                if origin==self.social_origin_facebook:
                    valid, data = FacebookSession.validate(token)
                    if valid:
                        id=data['id']
                        nombre=data['name']
                        email=data['email']
                        urlfoto=data['picture']['data']['url']
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(INVALID_TOKEN)
                elif origin==self.social_origin_google:
                    valid, data = GoogleSession.validate(token)
                    if valid:
                        id=data['sub']
                        nombre=data['name']
                        email=data['email']
                        urlfoto=data['picture']
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(INVALID_TOKEN)
                else:
                    return False, None, '', ErrorManager.get_error_by_code(INVALID_TOKEN)
                if self.user_model.objects.filter(**{self.social_id_field:id}).exists():
                    user_instance=self.user_model.objects.get(**{self.social_id_field:id})
                    if self.active_field_name==None or getattr(user_instance, self.active_field_name):
                        token=self._generate_token(user_instance, 0)
                        if token!=None:
                            return True, user_instance, token, ErrorManager.get_error_by_code(NO_ERROR)
                        else:
                            return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(INVALID_CREDENTIALS)
                elif self.user_model.objects.filter(**{self.login_id_field_name:email}).exists():
                    user_instance=self.user_model.objects.get(**{self.login_id_field_name:email})
                    if self.active_field_name==None or not getattr(user_instance, self.active_field_name):
                        setattr(user_instance, self.social_id_field, id)
                        setattr(user_instance, self.social_origin_field, origin)
                        if self.active_field_name!=None:
                            setattr(user_instance, self.active_field_name, True)
                        user_instance.save()
                        token=self._generate_token(user_instance, 0)
                        if token!=None:
                            return True, user_instance, token, ErrorManager.get_error_by_code(NO_ERROR)
                        else:
                            return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(USER_ALREADY_EXISTS)
                else:
                    user_instance=self.user_model(**{self.social_id_field:id, self.name_field_name:nombre, self.login_id_field_name:email, self.social_origin_field:origin})
                    if self.active_field_name!=None:
                        setattr(user_instance, self.active_field_name, True)
                    photo=self.download_photo(urlfoto)
                    fname=HashManager.getSHA1file(photo)
                    getattr(user_instance, self.photo_field_name).save(f'{fname}.jpg', photo, save=True)
                    user_instance.save()
                    token=self._generate_token(user_instance, 0)
                    if token!=None:
                        return True, user_instance, token, ErrorManager.get_error_by_code(NO_ERROR)
                    else:
                        return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
            except:
                return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)
        else:
            return False, None, '', ErrorManager.get_error_by_code(INTERNAL_ERROR)

    def actual_user_attr_getter(self, field_name):
        """ Generate a function that recive info, **kwargs and return the value of field_name of the actual user """
        def actual_user_attr_getter_func(info, model_instance, **kwargs):
            valid, user_instance, error_message=self.validate_access(info.context, 'all')
            sub_attrs=field_name.split('__')
            compare_object=user_instance
            for sub_attr in sub_attrs:
                compare_object=getattr(compare_object, sub_attr)
                if compare_object==None and sub_attr!=sub_attrs[-1]:
                    return None
            return compare_object
        return actual_user_attr_getter_func
    
    def build_access_level_validator(self, model_field='rol'):
        """ Buuild a validator that validate if a model_instance.model_field is in the actual_user's access level """
        def access_level_validator(info, model_instance, **kwargs):
            valid, user_instance, error_message=self.validate_access(info.context, 'all')
            if valid:
                sub_attrs=model_field.split('__')
                compare_object=model_instance
                for sub_attr in sub_attrs:
                    compare_object=getattr(compare_object, sub_attr)
                    if compare_object==None and sub_attr!=sub_attrs[-1]:
                        return False
                if compare_object in self.group_manager.get_modify_permissions(user_instance):
                    return True
                else:
                    return False
            else:
                return False
        return access_level_validator
        
    def actual_user_comparer(self, actual_user_field, operator, model_field=None, kwargs_key=None, default_value=None):
        """ Build a function that compare the value of actual_user.actual_user_field with the value of model.model_field or kwargs[kwargs_key] or default_value or default_value(info, model_instance, **kwargs) by operator: '=', '!=', '>', '<', '>=', '<=', 'in', 'not in', 'like', 'not like' """
        def actual_user_comparer_func(info, model_instance, **kwargs):
            valid, user_instance, error_message=self.validate_access(info.context, 'all')
            sub_attrs_user=actual_user_field.split('__')
            operator_function=OPERATIONS[operator]
            compare_user_object=user_instance
            for sub_attr_user in sub_attrs_user:
                compare_user_object=getattr(compare_user_object, sub_attr_user)
                if compare_user_object==None and sub_attr_user!=sub_attrs_user[-1]:
                    return False
            user_attr_value = compare_user_object
            if model_field==None and kwargs_key==None:
                if callable(default_value):
                    return operator_function(user_attr_value, default_value(info, model_instance, kwargs))
                else:
                    return operator_function(user_attr_value, default_value)
            if model_instance!=None and model_field!=None:
                if valid:
                    sub_attrs_instance=model_field.split('__')
                    compare_instance_object=model_instance
                    for sub_attr_instance in sub_attrs_instance:
                        compare_instance_object=getattr(compare_instance_object, sub_attr_instance)
                        if compare_instance_object==None and sub_attr_instance!=sub_attrs_instance[-1]:
                            return False
                    return operator_function(user_attr_value, compare_instance_object)
            if kwargs_key!=None and kwargs_key in kwargs.keys():
                if valid:
                    return operator_function(user_attr_value, kwargs[kwargs_key])
            return False
        return actual_user_comparer_func

    


# graphene imports
import graphene
# shared helpers
from django_graphbox.helpers.shared import *
# django imports
from django.db.models import Q

# query resolver builders
def build_field_by_id_resolver(self):
    def field_resolver_function(parent, info, **kwargs):
        # get model config by this path
        operation_name=info.operation.selection_set.selections[0].name.value.lower()
        config=self._models_by_op_name[operation_name]
        # get access group for validate access
        access_group=get_access_group('field_by_id', config)
        if self._session_manager!=None:
            valid, actual_user_instance, error=self._session_manager.validate_access(info.context, access_group)
        else:
            valid=True
        if valid:
            model=config.get('model')
            result=model.objects.get(id=kwargs.get('id'))
            valid_operation=True
            if 'field_by_id' in config['validators_by_operation']:
                valid_operation=evaluate_result(config['validators_by_operation']['field_by_id'], info, result, **kwargs)
            if valid_operation:
                return result
        return None
    return field_resolver_function

def build_field_list_resolver(self):
    def list_resolver_function(parent, info, **kwargs):
        operation_name=info.operation.selection_set.selections[0].name.value.lower()
        config=self._models_by_op_name[operation_name]
        pagination_length=config.get('pagination_length')
        pagination_style=config.get('pagination_style')
        paginated_type=config.get('paginated_type')
        external_filters=config.get('external_filters')
        internal_filters=config.get('internal_filters')
        filters_operator=config.get('filters_operator')
        query_object=None
        for filter_config in external_filters:
            param_value=kwargs.get(filter_config.get('param_name'))
            if param_value is not None:
                if query_object is None:
                    query_object=Q(**{filter_config.get('field_name'): param_value})
                else:
                    query_object.add(Q(**{filter_config.get('field_name'): param_value}), filters_operator)
        for filter_config in internal_filters:
            resolver_filter=filter_config.get('resolver_filter')
            on_return_none=filter_config.get('on_return_none')
            value_filter=resolver_filter(info, **kwargs)
            if value_filter==None:
                if on_return_none=='skip':
                    continue
                elif on_return_none=='set__isnull':
                    if query_object is None:
                        query_object=Q(**{f"{filter_config.get('field_name')}__isnull": True})
                    else:
                        query_object.add(Q(**{f"{filter_config.get('field_name')}__isnull": True}), filters_operator)
            else:
                if query_object is None:
                    query_object=Q(**{filter_config.get('field_name'): value_filter})
                else:
                    query_object.add(Q(**{filter_config.get('field_name'): value_filter}), filters_operator)
        if query_object is None:
            query_object=Q()
        # get access group for validate access
        access_group=get_access_group('list_field', config)
        if self._session_manager!=None:
            valid, actual_user_instance, error=self._session_manager.validate_access(info.context, access_group)
        else:
            valid=True
        if valid:
            model=config.get('model')
            if pagination_length == 0:
                return model.objects.filter(query_object)
            else:
                pagina=kwargs.get('page')
                inicio=(pagina*pagination_length)-pagination_length
                fin=inicio+pagination_length
                items=model.objects.filter(query_object)[inicio:fin]
                if pagination_style=='infinite':
                    return items
                else:
                    total_items=model.objects.filter(query_object).count()
                    total_pages=total_items//pagination_length
                    if total_items%pagination_length>0:
                        total_pages+=1
                    has_next_page = pagina<total_pages
                    has_previous_page = pagina>1
                    return paginated_type(items=items, has_next_page=has_next_page, has_previous_page=has_previous_page, total_pages=total_pages, total_items=total_items)
        return None
    return list_resolver_function

# filters_args getter

def get_filters_args(model_config):
    filters_args={}
    external_filters=model_config.get('external_filters')
    for filter_config in external_filters:
        filters_args[filter_config.get('param_name')]=filter_config.get('param_type')
    if model_config.get('pagination_length') != 0:
        filters_args['page']=graphene.Int(required=True)
    return filters_args

# return objects by pagination style

def get_return_object(model_config):
    filters_args=get_filters_args(model_config)
    if model_config.get('pagination_length')==0  or model_config.get('pagination_style') == 'infinite':
        return_object=graphene.List(model_config['type'], filters_args)
    elif model_config.get('pagination_length')>0 and model_config.get('pagination_style') == 'paginated':
        return_object=graphene.Field(model_config['paginated_type'], filters_args)
    else:
        raise Exception(f"Unknown pagination style {model_config.get('pagination_style')}")
    return return_object


# Dominant Access Group Getter

def get_access_group(operation, model_config):
    """ Get access group for operation based on model config 
    
    Args:
        operation (str): operation name
        model_config (dict): model config

    Returns:
        str: access group finaly selected for operation
    """
    if operation in model_config['access_by_operation']:
        return model_config['access_by_operation'][operation]
    return model_config['access_group']

# recursive logical expression evaluator for validators

def evaluate_result(operation, info, model_instance, **kwargs):
    """Evaluate result of operation validators and return a boolean result

    Args:
        operation (dict): operation validators config.
        info (dict): graphql.execution.base.ResolveInfo object.
        model_instance (object): model instance of operation.
        **kwargs (dict): kwargs input from graphql.
    Returns:
        bool: Result of evaluation of validators of operation config.
    """
    if 'validators' in operation:
        validators = operation['validators']
        if 'connector' in operation:
            connector = operation['connector']
        else:
            connector = 'AND'
        if connector == 'AND':
            result = True
            for validator in validators:
                if callable(validator):
                    result = result and validator(info, model_instance, **kwargs)
                else:
                    if validator==None:
                        raise Exception('Validator must be a callable or dict with validators and connector')
                    else:
                        result = result and evaluate_result(validator, info, model_instance, **kwargs)
            return result
        else:
            result = False
            for validator in validators:
                if callable(validator):
                    result = result or validator(info, model_instance, **kwargs)
                else:
                    if validator==None:
                        raise Exception('Validator must be a callable or dict with validators and connector')
                    else:
                        result = result or evaluate_result(validator, info, model_instance, **kwargs)
            return result
    else:
        return True




import hashlib

class HashManager:
    """ Hash manager class """
    @classmethod
    def getSHA1file(cls, file):
        """ Get SHA1 hash of file 
        
        Args:
            file (File): File to hash

        Returns:
            str: SHA1 hash of file
        """
        BLOCKSIZE = 65536
        hasher = hashlib.sha1()
        buf = file.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = file.read(BLOCKSIZE)
        return (hasher.hexdigest())


# graphene imports
import graphene
from graphene_django.types import DjangoObjectType
# error management
from .exceptions import ErrorManager, ErrorMsgType
# global constants
from .constants import *
# build helpers
from .helpers.mutations import *
from .helpers.queries import *
from .helpers.sessions import *

class SchemaBuilder:
    """ Class provides the functionality to build a GraphQL schema with basic operations: field_by_id, list_field, create_field, update_field and delete_field."""
    _models_config = {}
    _models_by_op_name = {}

    def __init__(self, session_manager=None):
        """Initialize the schema builder.

        Args:
            session_manager (SessionManager): Session manager to use.
        """
        self._session_manager = session_manager

    def add_model(self, model, exclude_fields=(), pagination_length=0, pagination_style='infinite', external_filters=[], internal_filters=[], filters_opeator=Q.AND, access_group=None, access_by_operation={}, validators_by_operation={}, internal_field_resolvers={}, exclude_fields_by_operation={}, save_as_password=[]):
        """Add model for build operations.

        Args:
            model (Model): Model to add to the schema.
            exclude_fields (tuple): Fields to exclude from the model type.
            pagination_length (int): Number of items to return in a paginated response. 0 means no pagination.
            pagination_style (str): Pagination style. Possible values are 'infinite' and 'paginated'.
            external_filters (list): Filters to apply to the model. 
            internal_filters (list): Internal filters to apply to the model. 
            filters_opeator (Q.AND, Q.OR): Operator to use for the filters.
            access_group (str): Access group to use for this model. This will be overriden by the access_by_operation.
            access_by_operation (dict): Dictionary with the operations to use for the access. {'operation': 'access_group', ...}
            validators_by_operation (dict): Dictionary with the validators to use for the access. {'operation': callable(info, model_instance, **kwargs), ...}
            internal_field_resolvers (dict): Dictionary with the internal field value resolvers on create_field and update_field operations
            save_as_password (list): List of fields to save as password.
        """
        #get the model name
        model_name = model.__name__
        #crreate the model type
        model_metaclass = type(f"Meta", (), {'model': model, 'exclude_fields': exclude_fields})
        model_type = type(f"{model_name}Type", (DjangoObjectType,), {'Meta': model_metaclass})
        #create paginated type
        if pagination_length > 0 and pagination_style == 'paginated':
            paginated_type = type(f"{model_name}PageType", (graphene.ObjectType,), {'items': graphene.List(model_type), 'page': graphene.Int(), 'has_next_page': graphene.Boolean(), 'has_previous_page': graphene.Boolean(), 'total_pages': graphene.Int(), 'total_items': graphene.Int()})
        else:
            paginated_type = None
        #make a new model config
        config = {
            'model': model,
            'name': model_name,
            'type': model_type,
            'pagination_length': pagination_length,
            'pagination_style': pagination_style,
            'paginated_type': paginated_type,
            'external_filters': external_filters,
            'internal_filters': internal_filters,
            'filters_operator': filters_opeator,
            'access_group': access_group,
            'access_by_operation': access_by_operation,
            'validators_by_operation': validators_by_operation,
            'internal_field_resolvers': internal_field_resolvers,
            'exclude_fields_by_operation': exclude_fields_by_operation,
            'save_as_password': save_as_password,
        }
        self._models_config[model_name]=config

    def build_schema_query(self):
        """ Build query class for the schema.

        Returns:
            graphene.ObjectType: Query class for the schema.
        """
        query_class= type("Query", (graphene.ObjectType,), {})
        for key in self._models_config.keys():
            model_config=self._models_config[key]
            object_name=model_config['name'].lower()
            # build field_by_id query
            field_by_id_resolver_function=build_field_by_id_resolver(self)
            self._models_by_op_name[object_name]=model_config
            setattr(query_class, object_name, graphene.Field(model_config['type'], id=graphene.ID(required=True)))
            setattr(query_class, f'resolve_{object_name}', field_by_id_resolver_function)
            # build list_field query
            field_list_resolver_function=build_field_list_resolver(self)
            self._models_by_op_name['all' + object_name]=model_config
            return_object=get_return_object(model_config)
            setattr(query_class, f"all_{object_name}", return_object)
            setattr(query_class, f'resolve_all_{object_name}',field_list_resolver_function)
        return query_class
    
    def build_schema_mutation(self):
        """Build mutations class for the schema. 
        
        Returns:
            graphene.ObjectType: Mutations class for the schema.
        """
        mutation_class= type("Mutation", (), {})
        for key in self._models_config.keys():
            model_config=self._models_config[key]
            # create the create mutation
            mutate_create_function=build_mutate_for_create(self)
            # get fields to ignore on arguments
            fields_to_ignore=get_fields_to_ignore(model_config, 'create_field')
            # build argumants class
            arguments_create=create_arguments_class(model_config['model'], fields_to_ignore)
            create_mutation=type("Create"+model_config['name'], (graphene.Mutation,), {"estado":graphene.Boolean(),model_config['name'].lower(): graphene.Field(model_config['type']), "error":graphene.Field(ErrorMsgType), 'Arguments': arguments_create, 'mutate': mutate_create_function})
            setattr(mutation_class, f"create_{model_config['name'].lower()}", create_mutation.Field())
            self._models_by_op_name['create' + model_config['name'].lower()]=model_config
            # create the update mutation
            mutate_update_function=build_mutate_for_update(self)
            # get fields to omit
            fields_to_ignore=get_fields_to_ignore(model_config, 'update_field')
            # build argumants class
            arguments_update=update_arguments_class(model_config['model'], fields_to_ignore, model_config.get('save_as_password'))
            update_mutation=type("Update"+model_config['name'], (graphene.Mutation,), {"estado":graphene.Boolean(),model_config['name'].lower(): graphene.Field(model_config['type']), "error":graphene.Field(ErrorMsgType), 'Arguments': arguments_update, 'mutate': mutate_update_function})
            setattr(mutation_class, f"update_{model_config['name'].lower()}", update_mutation.Field())
            self._models_by_op_name['update' + model_config['name'].lower()]=model_config
            # create the delete mutation
            mutate_delete_function=build_mutate_for_delete(self)
            # build argumants class
            delete_arguments=delete_arguments_class()
            delete_mutation=type("Delete"+model_config['name'], (graphene.Mutation,), {"estado":graphene.Boolean(), "error":graphene.Field(ErrorMsgType), 'Arguments': delete_arguments, 'mutate': mutate_delete_function})
            setattr(mutation_class, f"delete_{model_config['name'].lower()}", delete_mutation.Field())
            self._models_by_op_name['delete' + model_config['name'].lower()]=model_config
        return mutation_class

    def build_session_schema(self):
        """ Build the session mutations and queries

        Returns:
            tuple: (session_query_class, session_mutation_class)
        """
        if self._session_manager!=None:
            if self._session_manager.user_model.__name__ in self._models_config.keys():
                config_login_model=self._models_config[self._session_manager.user_model.__name__]
            else:
                #get the model name
                model_name = self._session_manager.user_model.__name__
                #crreate the model type
                model_metaclass = type(f"Meta", (), {'model': self._session_manager.user_model, 'excld_fields': (self._session_manager.password_field_name,)})
                model_type = type(f"{model_name}Type", (DjangoObjectType,), {'Meta': model_metaclass})
                config_login_model = {
                    'model': self._session_manager.user_model,
                    'name': model_name,
                    'type': model_type,
                    'pagination_length': 0,
                    'filters': {},
                    'filters_operator': Q.AND,
                    'access_group': None,
                    'access_by_operation': {}
                }
                self._models_config[model_name]=config_login_model
            #build Login Mutation
            mutation_class= type("Mutation", (), {})
            arguments_class= type("Arguments", (), {'permanent':graphene.Boolean(required=True)})
            setattr(arguments_class, self._session_manager.login_id_field_name, graphene.String(required=True))
            setattr(arguments_class, self._session_manager.password_field_name, graphene.String(required=True))
            login_mutate_function=build_mutate_for_login(self)
            login_mutation=type("Login", (graphene.Mutation,), {'estado': graphene.Boolean(), config_login_model['name'].lower(): graphene.Field(config_login_model['type']), 'token':graphene.String(), 'error': graphene.Field(ErrorMsgType), 'Arguments': arguments_class, 'mutate': login_mutate_function})
            setattr(mutation_class, "login", login_mutation.Field())
            self._models_by_op_name['login']=config_login_model
            # build Social Login Mutation
            if self._session_manager.use_social_session:
                arguments_class= type("Arguments", (), {'token':graphene.String(required=True), 'origin':graphene.String(required=True)})
                social_login_mutate_function=build_mutate_for_social_login(self)
                social_login_mutation=type("SocialLogin", (graphene.Mutation,), {'estado': graphene.Boolean(), config_login_model['name'].lower(): graphene.Field(config_login_model['type']), 'token':graphene.String(), 'error': graphene.Field(ErrorMsgType), 'Arguments': arguments_class, 'mutate': social_login_mutate_function})
                setattr(mutation_class, "social_login", social_login_mutation.Field())
            # build actual_user query
            query_class= type("Query", (graphene.ObjectType,), {})
            actual_user_function=build_actual_user_resolver(self)
            setattr(query_class, "actual_user", graphene.Field(config_login_model['type']))
            setattr(query_class, "resolve_actual_user", actual_user_function)
            self._models_by_op_name['actualuser']=config_login_model
            return query_class, mutation_class
        else:
            assert False, "Session Manager not defined"